#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include "ptp.h"
#include "ptpcommon.h"
#include "ptpraw.h"
#include "ptpclient.h"

#define	PTP_RECV_ERR_SLEEP	500000		// sleeps 0.5 seconds upon a major PTP receive error (from PTP handler)

// workaround since pthreads can't start on a class method
void *ptp_thread_stub(void *object)
{
	PTPclient	*my_this = (PTPclient *)object;

	my_this->thread_run();

	return(object);	// this is never reached
}

PTPclient::PTPclient()
{
	this->is_init = 0;
	this->state = PTP_STATE_STOPPED;
	this->error_code = PTP_ERR_NONE;

	this->ptp_address_table = NULL;
	this->ptp_address_entries = 0;
	this->ptp_bandwidth_ok = 0;
	this->ptp_min_bandwidth = 0;
}

PTPclient::~PTPclient()
{
	this->shutdown();
}

int PTPclient::get_state(int *error_code)
{
	if(error_code)
		*error_code = this->error_code;

	return(this->state);
}

int PTPclient::init(const char *device, const char *remote, __u16 ptp_port)
{
	int		if_fd;
	struct ifreq	if_req;
	struct in_addr	inaddr, inaddr_remote;
	struct hostent	*he;

	if(this->is_init)
		this->shutdown();

	// try to determine address to bind to
	if_fd = socket(AF_INET, SOCK_DGRAM, 0);
	if(if_fd) {
		strcpy(if_req.ifr_name, device);
		if_req.ifr_addr.sa_family = AF_INET;
		if(ioctl(if_fd, SIOCGIFADDR, &if_req) >= 0)
			this->device = ((struct sockaddr_in *)&if_req.ifr_addr)->sin_addr.s_addr;
		else {
			this->error_code = PTP_ERR_NOSUCHDEVICE;
			close(if_fd);
			return(0);
		}

		close(if_fd);
	}
	else {
		this->error_code = PTP_ERR_CREATESOCKET;
		return(0);
	}

	// get remote address
	// try dotted style
	if(!inet_aton(remote, &inaddr_remote)) {
		// use DNS lookup
		if((he = gethostbyname(remote)) == NULL) {
			this->error_code = PTP_ERR_HOSTLOOKUP;
			return(0);
		}
		memcpy((char *)&inaddr_remote, he->h_addr, he->h_length);
	}

	// initialize lowlevel PTP packet transportation code
	inaddr.s_addr = this->device;
	if(!this->ptp_raw.init(inaddr, inaddr_remote, ptp_port)) {
		this->error_code = PTP_ERR_INITRAW;
		return(0);
	}

	pthread_attr_init(&this->ptp_attr);

	if(pthread_create(&this->ptp_thread, &this->ptp_attr, ptp_thread_stub, (void *)this)) {
		// error creating thread
		this->error_code = PTP_ERR_CREATETHREAD;
		return(0);
	}

	this->state = PTP_STATE_RUNNING;
	this->is_init = 1;
	return(1);
}

void PTPclient::shutdown(void)
{
	ptpAddressTable	*temp_ad, *temp_ad1;

	if(this->is_init) {
		this->is_init = 0;
		pthread_cancel(this->ptp_thread);
		pthread_attr_destroy(&this->ptp_attr);
	}

	if(this->ptp_address_table) {
		temp_ad = ptp_address_table;
		while(temp_ad) {
			temp_ad1 = temp_ad->next;
			delete(temp_ad);
			temp_ad = temp_ad1;
		}

		this->ptp_address_table = NULL;
	}

	this->ptp_address_entries = 0;
	this->ptp_bandwidth_ok = 0;
	this->ptp_min_bandwidth = 0;

	this->state = PTP_STATE_STOPPED;
}

void PTPclient::thread_run(void)
{
	int		len;
	unsigned char	*packet;
	struct iphdr	*ip_hdr;
	struct ptphdr	*ptp_hdr;

	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	while(1) {
		// wait for a PTP packet
		len = this->ptp_raw.recv_ptp(-1, &packet);
		if(len) {
			this->error_code = PTP_ERR_NONE;

			ip_hdr = (struct iphdr*)packet;
			ptp_hdr = (struct ptphdr*)(packet + ip_hdr->ihl*4);

			_DMSG1("[thread_run(): recv'd PTP packet type %d]\n", ptp_hdr->type);

			// eliminate response-packets (compare=0, echo=1)
			if(ptp_hdr->compare || !ptp_hdr->echo) {
				// sanity check
				if((unsigned)len < (sizeof(struct ptphdr) + ip_hdr->ihl*4 + ptp_hdr->count*ptp_hdr->length*2)) {
					_DMSG1("[thread_run(): PTP packet ds_count/PDU_len mismatch! (packet_len: %d)]\n", len);
					_DMSG3("[thread_run(): ds_count: %d, real ds_length: %d, PDU_len: %d]\n", ptp_hdr->count, ptp_hdr->length*2, len - sizeof(struct ptphdr) - ip_hdr->ihl*4);
				}
				else {
					// evaluate packet
					switch(ptp_hdr->type) {
					case PTP_TYPE_IFMTU:
						this->process_ifmtu(packet, len);
						break;

					case PTP_TYPE_IFSPEED:
						this->process_ifspeed(packet, len);
						break;

					case PTP_TYPE_ADDR_IFSPEED:
						if(!ptp_hdr->compare)
							this->process_addr_ifspeed(packet, len);
						break;

					case PTP_TYPE_ADDR_TIME_IFOCTETS:
						if(!ptp_hdr->compare)
							this->process_addr_time_ifoctets(packet, len);
						break;

					case PTP_TYPE_ADDR_TIME_IFSPEED_IFOCTETS:
						if(!ptp_hdr->compare)
							this->process_addr_time_ifspeed_ifoctets(packet, len);
						break;
					}
				}
			}
#ifdef __DEBUG
			else
				_DMSG0("[thread_run(): packet had c=0 e=1 -> nonsense]\n");
#endif

			delete(packet);
		}
		else {
			this->error_code = PTP_ERR_RECVRAW;

			// we really need to sleep a bit to avoid polling (in case the PTP handler is broken)
			usleep(PTP_RECV_ERR_SLEEP);
		}
	}
}

// simple ifMTU packet handling
void PTPclient::process_ifmtu(const unsigned char *packet, int len)
{
	int		packet_len;
	unsigned char	*packet_out, *data;
	struct iphdr	*ip_hdr;
	struct ptphdr	*ptp_hdr;
	__u8		count;
	__u32		worst, worst_n, *temp_ptr, data_out[2];

	_DMSG0("[process_ifmtu(): processing IFMTU packet]\n");
	// check packet for consistency, send packet containing the worst MTU back to sender
	ip_hdr = (struct iphdr*)packet;
	ptp_hdr = (struct ptphdr*)(packet + ip_hdr->ihl*4);
	data = (unsigned char *)((unsigned char *)ptp_hdr + sizeof(struct ptphdr));

	// check dataset length
	if(ptp_hdr->length % 2) {
		_DMSG1("[process_ifmtu(): length mismatch (%d)]\n", ptp_hdr->length);
		return;
	}

	// check dataset count
	if(ptp_hdr->count < 1) {
		_DMSG1("[process_ifmtu(): ds_count mismatch (%d)]\n", ptp_hdr->count);
		return;
	}

	if(ptp_hdr->compare) {
		// the easy case: send the one and only dataset back
		if(!ptp_hdr->echo) {
			_DMSG0("[process_ifmtu(): preparing response to c=1, e=0 packet]\n");
			data_out[0] = *((__u32 *)data);
			data_out[1] = 0;
			packet_out = this->ptp_raw.assemble_ptp(PTP_DEFAULT_TTL, ip_hdr->saddr, PTP_DEFAULT_TTL, 0, 1, 2, ptp_hdr->count, PTP_TYPE_IFMTU, sizeof(__u32) * 2, sizeof(__u32) * 2, (unsigned char *)data_out, &packet_len);
		}
		else {
			_DMSG0("[process_ifmtu(): preparing response to c=1, e=1 packet]\n");
			data_out[0] = *((__u32 *)data);
			data_out[1] = 0;
			packet_out = this->ptp_raw.assemble_ptp(PTP_DEFAULT_TTL, ip_hdr->saddr, PTP_DEFAULT_TTL, 0, 1, 2, ptp_hdr->count, PTP_TYPE_IFMTU, sizeof(__u32) * 2, sizeof(__u32) * 2, (unsigned char *)data_out, &packet_len);
		}
	}
	else {
		// locate the worst MTU and send it back
		_DMSG0("[process_ifmtu(): preparing response to c=0 packet]\n");
		temp_ptr = (__u32 *)data;
		worst = ntohl(*temp_ptr);
		temp_ptr++;

		count = 1;
		while(count < ptp_hdr->count) {
			if(ntohl(*temp_ptr) < worst)
				worst = ntohl(*temp_ptr);

			count++;
			temp_ptr++;
		}

		worst_n = htonl(worst);
		data_out[0] = worst_n;
		data_out[1] = 0;
		packet_out = this->ptp_raw.assemble_ptp(PTP_DEFAULT_TTL, ip_hdr->saddr, PTP_DEFAULT_TTL, 0, 1, 2, PTP_DEFAULT_TTL - ip_hdr->ttl - ptp_hdr->count, PTP_TYPE_IFMTU, sizeof(__u32) * 2, sizeof(__u32) * 2, (unsigned char *)data_out, &packet_len);
	}

	if(packet_out) {
		_DMSG0("[process_ifmtu(): sending response]\n");
		if(!this->ptp_raw.send_ptp(packet_len, packet_out)) {
			_DMSG0("[process_ifmtu(): send failed]\n");
			this->error_code = PTP_ERR_SENDRAW;
		}

		delete(packet_out);
	}
}

// simple ifSpeed packet handling (just like ifMTU)
void PTPclient::process_ifspeed(const unsigned char *packet, int len)
{
	int		packet_len;
	unsigned char	*packet_out, *data;
	struct iphdr	*ip_hdr;
	struct ptphdr	*ptp_hdr;
	__u8		count;
	__u32		worst, worst_n, *temp_ptr, data_out[2];

	_DMSG0("[process_ifspeed(): processing IFSPEED packet]\n");
	// check packet for consistency, send packet containing the worst speed back to sender
	ip_hdr = (struct iphdr*)packet;
	ptp_hdr = (struct ptphdr*)(packet + ip_hdr->ihl*4);
	data = (unsigned char *)((unsigned char *)ptp_hdr + sizeof(struct ptphdr));

	// check dataset length
	if(ptp_hdr->length % 2) {
		_DMSG1("[process_ifspeed(): length mismatch (%d)]\n", ptp_hdr->length);
		return;
	}

	// check dataset count
	if(ptp_hdr->count < 1) {
		_DMSG1("[process_ifspeed(): ds_count mismatch (%d)]\n", ptp_hdr->count);
		return;
	}

	if(ptp_hdr->compare) {
		// the easy case: send the one and only dataset back
		if(!ptp_hdr->echo) {
			_DMSG0("[process_ifspeed(): preparing response to c=1, e=0 packet]\n");
			data_out[0] = *((__u32 *)data);
			data_out[1] = 0;
			packet_out = this->ptp_raw.assemble_ptp(PTP_DEFAULT_TTL, ip_hdr->saddr, PTP_DEFAULT_TTL, 0, 1, 2, ptp_hdr->count, PTP_TYPE_IFSPEED, sizeof(__u32) * 2, sizeof(__u32) * 2, (unsigned char *)data_out, &packet_len);
		}
		else {
			_DMSG0("[process_ifspeed(): preparing response to c=1, e=1 packet]\n");
			data_out[0] = *((__u32 *)data);
			data_out[1] = 0;
			packet_out = this->ptp_raw.assemble_ptp(PTP_DEFAULT_TTL, ip_hdr->saddr, PTP_DEFAULT_TTL, 0, 1, 2, ptp_hdr->count, PTP_TYPE_IFSPEED, sizeof(__u32) * 2, sizeof(__u32) * 2, (unsigned char *)data_out, &packet_len);
		}
	}
	else {
		// locate the worst speed and send it back
		_DMSG0("[process_ifspeed(): preparing response to c=0 packet]\n");
		temp_ptr = (__u32 *)data;
		worst = ntohl(*temp_ptr);
		temp_ptr++;

		count = 1;
		while(count < ptp_hdr->count) {
			if(ntohl(*temp_ptr) < worst)
				worst = ntohl(*temp_ptr);

			count++;
			temp_ptr++;
		}

		worst_n = htonl(worst);
		data_out[0] = worst_n;
		data_out[1] = 0;
		packet_out = this->ptp_raw.assemble_ptp(PTP_DEFAULT_TTL, ip_hdr->saddr, PTP_DEFAULT_TTL, 0, 1, 2, PTP_DEFAULT_TTL - ip_hdr->ttl - ptp_hdr->count, PTP_TYPE_IFSPEED, sizeof(__u32) * 2, sizeof(__u32) * 2, (unsigned char *)data_out, &packet_len);
	}

	if(packet_out) {
		_DMSG0("[process_ifspeed(): sending response]\n");
		if(!this->ptp_raw.send_ptp(packet_len, packet_out)) {
			_DMSG0("[process_ifspeed(): send failed]\n");
			this->error_code = PTP_ERR_SENDRAW;
		}

		delete(packet_out);
	}
}

// updates or creates the internal address table
void PTPclient::update_address_table(int skip_fields, int ds_count, const __u32 *data)
{
	int		n, replace_table = 0;
	const __u32	*temp_ptr;
	ptpAddressTable	*created_ad = NULL, *temp_ad = NULL, *temp_ad1, *new_ad;

	// addr, [timestamp,] ifSpeed [,ifOctets]; addr, [timestamp,] ifSpeed [,ifOctets]; ...
//	_DMSG0("[update_address_table(): creating new table]\n");
	temp_ptr = data;
	for(n = 0; n < ds_count; n++) {
		new_ad = new(ptpAddressTable);
		new_ad->next = NULL;
		new_ad->address = *(temp_ptr++);

		if(skip_fields)
			temp_ptr++;	// we were called from a content-type 4 instead of 2, skip timestamp

		new_ad->ifspeed = ntohl(*(temp_ptr++));

		if(skip_fields)
			temp_ptr++;	// we were called from a content-type 4 instead of 2, skip ifOctets

		new_ad->time_old = new_ad->time_new = 0;
		new_ad->ifoctets_old = new_ad->ifoctets_new = 0;

		if(temp_ad)
			temp_ad->next = new_ad;
		else
			created_ad = new_ad;

		temp_ad = new_ad;
	}

	// do we need to create a new table?
	if(this->ptp_address_entries == ds_count) {
		// check if table matches new one
		temp_ad = this->ptp_address_table;
		temp_ad1 = created_ad;
		while(!replace_table && temp_ad && temp_ad1) {
			if(temp_ad->address != temp_ad1->address)
				replace_table = 1;
			else {
				temp_ad = temp_ad->next;
				temp_ad1 = temp_ad1->next;
			}
		}

		if(temp_ad || temp_ad1)
			replace_table = 1;	// weird..
	}
	else
		replace_table = 1;

	if(replace_table) {
		// wipe old list
		_DMSG0("[update_address_table(): addresses changed, replacing old table]\n");
		temp_ad = this->ptp_address_table;
		while(temp_ad) {
			temp_ad1 = temp_ad->next;
			delete(temp_ad);
			temp_ad = temp_ad1;
		}

		this->ptp_address_table = created_ad;
		this->ptp_address_entries = ds_count;
		this->ptp_bandwidth_ok = 0;
	}
	else {
		// wipe new list
//		_DMSG0("[update_address_table(): addresses did not change, dropping newly created table]\n");
		temp_ad = created_ad;
		while(temp_ad) {
			temp_ad1 = temp_ad->next;
			delete(temp_ad);
			temp_ad = temp_ad1;
		}
	}
}

// addr_ifSpeed packet handling (this one is used to build an address table)
void PTPclient::process_addr_ifspeed(const unsigned char *packet, int len)
{
	unsigned char	*data;
	struct iphdr	*ip_hdr;
	struct ptphdr	*ptp_hdr;

	// check packet for consistency, build address table
	ip_hdr = (struct iphdr*)packet;
	ptp_hdr = (struct ptphdr*)(packet + ip_hdr->ihl*4);
	data = (unsigned char *)((unsigned char *)ptp_hdr + sizeof(struct ptphdr));

	// check dataset length
	if(ptp_hdr->length % 4)
		return;

	// check dataset count
	if(ptp_hdr->count < 1)
		return;

	// finally generate/update internal address table
	this->update_address_table(0, ptp_hdr->count, (__u32 *)data);
}

// update timestamp and ifoctet values in internal address table,
// send current-bandwidth-msg back to sender if possible, or route-change-msg back to source when appropriate
void PTPclient::handle_bandwidth_measurement(int ds_count, const __u32 *data, __u8 ptp_type, __u32 src_addr, __u8 ttl)
{
	int		n = 0, route_changed = 0, packet_len;
	unsigned char	*packet_out;
	const __u32	*temp_ptr;
	__u32		max_ifspeed = 0, bandwidth, interval, max_interval = 0, min_bandwidth = 0xffffffff, min_bw_interval, traffic, ifspeed, temp_rci;
	struct ptpmsg	ptp_msg;
	ptpAddressTable	*temp_ad;

	// update values in table
	if(ds_count != this->ptp_address_entries)
		route_changed = 1;
	else {
		// addr, timestamp, [ifSpeed,] ifOctets; addr, timestamp, [ifSpeed,] ifOctets; ...
		n = 0;
		temp_ad = this->ptp_address_table;
		temp_ptr = data;
		while(!route_changed && temp_ad && (n < ds_count)) {
			if(temp_ad->address == *(temp_ptr++)) {
				temp_ad->time_old = temp_ad->time_new;
				temp_ad->time_new = ntohl(*(temp_ptr++));

				if(ptp_type != PTP_TYPE_ADDR_TIME_IFOCTETS)
					temp_ptr++;	// we were called from a content-type 4 instead of 3, skip ifSpeed

				temp_ad->ifoctets_old = temp_ad->ifoctets_new;
				temp_ad->ifoctets_new = ntohl(*(temp_ptr++));
			}
			else
				route_changed = 1;

			temp_ad = temp_ad->next;
			n++;
		}

		if(n != ds_count)
			route_changed = 1;	// one more sanity check...
	}

	if(route_changed) {
		// send route-change-msg back to sender
		_DMSG0("[handle_bandwidth_measurement(): route has changed, sending route-changed-err]\n");

		/* route = 1; counter = 0; interval = 0 */
		temp_rci = 0; temp_rci = temp_rci | ROUTE_CHANGE;
		ptp_msg.rci = htonl(temp_rci);

		ptp_msg.mib2_ifspeed = ptp_msg.traffic = 0;

		packet_out = this->ptp_raw.assemble_ptp(PTP_DEFAULT_TTL, src_addr, PTP_DEFAULT_TTL, 0, 1, 6, 1, ptp_type, sizeof(struct ptpmsg), sizeof(struct ptpmsg), (unsigned char *)&ptp_msg, &packet_len);
		if(packet_out) {
			if(!this->ptp_raw.send_ptp(packet_len, packet_out))
				this->error_code = PTP_ERR_SENDRAW;

			delete(packet_out);
		}
		return;
	}

	// calculate bandwidth (if possible already)
	if(this->ptp_bandwidth_ok) {
		// calculate minimum bandwidth and remember maximum ifspeed
		_DMSG0("[handle_bandwidth_measurement(): calculating bandwidth]\n");
		temp_ad = this->ptp_address_table;
		while(temp_ad) {
			if(temp_ad->ifspeed > max_ifspeed)
				max_ifspeed = temp_ad->ifspeed;

			//_DMSG2("[handle_bandwidth_measurement(): time_old: %u, time_new: %u]\n", temp_ad->time_old, temp_ad->time_new);
			interval = temp_ad->time_new - temp_ad->time_old;
			if(interval > max_interval)
				max_interval = interval;

			//_DMSG2("[handle_bandwidth_measurement(): ifoctets_old: %u, ifoctets_new: %u]\n", temp_ad->ifoctets_old, temp_ad->ifoctets_new);
			bandwidth = temp_ad->ifspeed - (temp_ad->ifoctets_new - temp_ad->ifoctets_old) / interval;

			if(bandwidth < min_bandwidth) {
				//_DMSG1("[handle_bandwidth_measurement(): min_bw changed to %u]\n", bandwidth);
				min_bandwidth = bandwidth;
				min_bw_interval = interval;
				traffic = temp_ad->ifoctets_new - temp_ad->ifoctets_old;
				ifspeed = temp_ad->ifspeed;
			}

			temp_ad = temp_ad->next;
		}

		if(max_ifspeed == 0)
			max_ifspeed = 1;	// who knows...

		if(max_interval <= 0x3fffffff) {
			// Create the normal feedbackmsg
			_DMSG0("[handle_bandwidth_measurement(): sending bandwidth]\n");

			/* route = 0; counter = 0; interval = interval */
			temp_rci = 0; temp_rci = temp_rci | min_bw_interval;
			ptp_msg.rci = htonl(temp_rci);

			ptp_msg.mib2_ifspeed = ifspeed;
			ptp_msg.traffic = traffic;

			packet_out = this->ptp_raw.assemble_ptp(PTP_DEFAULT_TTL, src_addr, PTP_DEFAULT_TTL, 0, 1, 6, 1, ptp_type, sizeof(struct ptpmsg), sizeof(struct ptpmsg), (unsigned char *) &ptp_msg, &packet_len);
			if(packet_out) {
				if(!this->ptp_raw.send_ptp(packet_len, packet_out)) {
					_DMSG0("[handle_bandwidth_measurement(): error sending packet]\n");
					this->error_code = PTP_ERR_SENDRAW;
				}
				delete(packet_out);
			}
			else _DMSG0("[handle_bandwidth_measurement(): error assembling packet]\n");

			this->ptp_min_bandwidth = min_bandwidth;
		}
		else {
			// send counter-overflow-error
			_DMSG2("[handle_bandwidth_measurement(): counter overflow (max_int: %d, max_ifspd: %d]\n", max_interval, max_ifspeed);

			/* route = 0; counter = 1; interval = max interval */
			temp_rci = 0;
			temp_rci = temp_rci | htonl(0xffffffff / (4 * max_ifspeed));
			temp_rci = temp_rci | COUNTER_OVERFLOW;
			ptp_msg.rci = htonl(temp_rci);

			ptp_msg.mib2_ifspeed = ptp_msg.traffic = 0;

			packet_out = this->ptp_raw.assemble_ptp(PTP_DEFAULT_TTL, src_addr, PTP_DEFAULT_TTL, 0, 1, 6, 1, ptp_type, sizeof(struct ptpmsg), sizeof(struct ptpmsg), (unsigned char *)&ptp_msg, &packet_len);
			if(packet_out) {
				if(!this->ptp_raw.send_ptp(packet_len, packet_out))
					this->error_code = PTP_ERR_SENDRAW;

				delete(packet_out);
			}
		}
	}
	else {
		_DMSG0("[handle_bandwidth_measurement(): not enough packets yet to calculate bandwidth]\n");
		this->ptp_bandwidth_ok = 1;	// we got enough data to take bandwidth next time
	}
}

// addr_time_ifOctets packet handling (this one is used to measure current bandwidth)
void PTPclient::process_addr_time_ifoctets(const unsigned char *packet, int len)
{
	unsigned char	*data;
	struct iphdr	*ip_hdr;
	struct ptphdr	*ptp_hdr;

	// check packet for consistency, add entries to table, send back bandwidth measurement (if possible)
	ip_hdr = (struct iphdr*)packet;
	ptp_hdr = (struct ptphdr*)(packet + ip_hdr->ihl*4);
	data = (unsigned char *)((unsigned char *)ptp_hdr + sizeof(struct ptphdr));

	// check dataset length
	if(ptp_hdr->length % 6)
		return;

	// check dataset count
	if(ptp_hdr->count < 1)
		return;

	// finally process packet data
	this->handle_bandwidth_measurement(ptp_hdr->count, (__u32 *)data, PTP_TYPE_ADDR_TIME_IFOCTETS, ip_hdr->saddr, ip_hdr->ttl);
}

// addr_time_ifSpeed_ifOctets packet handling (this one is used to build address table and measure current bandwidth)
void PTPclient::process_addr_time_ifspeed_ifoctets(const unsigned char *packet, int len)
{
	unsigned char	*data;
	struct iphdr	*ip_hdr;
	struct ptphdr	*ptp_hdr;

	// check packet for consistency, add entries to table, send back bandwidth measurement (if possible)
	ip_hdr = (struct iphdr*)packet;
	ptp_hdr = (struct ptphdr*)(packet + ip_hdr->ihl*4);
	data = (unsigned char *)((unsigned char *)ptp_hdr + sizeof(struct ptphdr));

	// check dataset length
	if(ptp_hdr->length % 8)
		return;

	// check dataset count
	if(ptp_hdr->count < 1)
		return;

	// finally generate/update internal address table and process packet data
	this->update_address_table(1, ptp_hdr->count, (__u32 *)data);
	this->handle_bandwidth_measurement(ptp_hdr->count, (__u32 *)data, PTP_TYPE_ADDR_TIME_IFSPEED_IFOCTETS, ip_hdr->saddr, ip_hdr->ttl);
}
