#include <stdio.h>
#include <unistd.h>
#include <pthread.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/times.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <net/if.h>
#include <netinet/ip.h>
#include <sys/ioctl.h>
#include "ptp.h"
#include "ptpcommon.h"
#include "ptpraw.h"
#include "ptpserver.h"

#define	PTP_RAWINIT_DELAY	100000	// 0.1secs sleep after raw-module init (unix domain sockets need a short time here)

// workaround since pthreads can't start on a class method
void *ptp_thread_stub(void *object)
{
	PTPserver	*my_this = (PTPserver *)object;

	my_this->thread_run();

	return(object);	// this is never reached
}

PTPserver::PTPserver()
{
	this->is_init = 0;
	this->running = 0;
	this->error_code = PTP_BANDWIDTH_ERR_NONE;
	this->error = PTP_BANDWIDTH_ERR_STOPPED;

	pthread_mutex_init(&this->ptp_mtx, NULL);
}

PTPserver::~PTPserver()
{
	this->stop_availbw_detection();

	pthread_mutex_destroy(&this->ptp_mtx);
}

int PTPserver::convert_to_inaddr(const char *host, struct in_addr *address)
{
	struct hostent	*he;

	// try dotted style
	if(inet_aton(host, address))
		return(1);

	// use DNS lookup
	if((he = gethostbyname(host)) == NULL)
		return(0);

	memcpy((char *)address, he->h_addr, he->h_length);
	return(1);
}

int PTPserver::init(const char *device, const char *destination, long rtt, __u16 ptp_port)
{
	int		if_fd;
	struct ifreq	if_req;
	struct in_addr	inaddr;

	this->stop_availbw_detection();

	this->rtt = rtt;
	this->ptp_port = ptp_port;

	if(!this->convert_to_inaddr(destination, &this->destination)) {
		this->error_code = PTP_ERR_HOSTLOOKUP;
		return(0);
	}

	// try to determine address to bind to
	if_fd = socket(AF_INET, SOCK_DGRAM, 0);
	if(if_fd) {
		strcpy(if_req.ifr_name, device);
		if_req.ifr_addr.sa_family = AF_INET;
		if(ioctl(if_fd, SIOCGIFADDR, &if_req) >= 0)
			this->device = ((struct sockaddr_in *)&if_req.ifr_addr)->sin_addr.s_addr;
		else {
			this->error_code = PTP_ERR_NOSUCHDEVICE;
			return(0);
		}

		close(if_fd);
	}
	else {
		this->error_code = PTP_ERR_CREATESOCKET;
		return(0);
	}

	// initialize lowlevel PTP packet transportation code
	inaddr.s_addr = this->device;
	if(!this->ptp_raw.init(inaddr, this->destination, this->ptp_port)) {
		this->error_code = PTP_ERR_INITRAW;
		return(0);
	}

	usleep(PTP_RAWINIT_DELAY);
	this->error_code = PTP_ERR_NONE;
	this->is_init = 1;
	return(1);
}

__u32 PTPserver::get_path_mtu(int compare, int echo, __u32 min_mtu, __u8 expected_hops, __u8 *missing)
{
	int		len, data_len;
	__u32		mtu, min_mtu_n;
	unsigned char	*packet, *padding;
	struct iphdr	*ip_hdr;
	struct ptphdr	*ptp_hdr;

	pthread_mutex_lock(&this->ptp_mtx);
	this->error_code = PTP_ERR_NONE;

	if(!compare && !echo) {
		if(expected_hops > 0) {
			padding = new(unsigned char[expected_hops * 4]);
			memset(padding, 0, expected_hops * 4);
		}
		else
			padding = NULL;

		packet = this->ptp_raw.assemble_ptp(PTP_DEFAULT_TTL, this->destination.s_addr, PTP_DEFAULT_TTL, compare, echo, 2,0, PTP_TYPE_IFMTU, expected_hops * 4, 0, padding, &len);

		if(padding)
			delete(padding);

		if(packet) {
			if(!this->ptp_raw.send_ptp(len, packet)) {
				delete(packet);
				this->error_code = PTP_ERR_SENDRAW;
				pthread_mutex_unlock(&this->ptp_mtx);
				return(0);
			}

			delete(packet);
		}
		else {
			this->error_code = PTP_ERR_CREATERAW;
			pthread_mutex_unlock(&this->ptp_mtx);
			return(0);
		}
	}
	else if(compare) {
		min_mtu_n = htonl(min_mtu);
		packet = this->ptp_raw.assemble_ptp(PTP_DEFAULT_TTL, this->destination.s_addr, PTP_DEFAULT_TTL, compare, echo, 2,1, PTP_TYPE_IFMTU, 4, 4, (unsigned char *)&min_mtu_n, &len);
		if(packet) {
			if(!this->ptp_raw.send_ptp(len, packet)) {
				delete(packet);
				this->error_code = PTP_ERR_SENDRAW;
				pthread_mutex_unlock(&this->ptp_mtx);
				return(0);
			}

			delete(packet);
		}
		else {
			this->error_code = PTP_ERR_CREATERAW;
			pthread_mutex_unlock(&this->ptp_mtx);
			return(0);
		}
	}
	else {
		this->error_code = PTP_ERR_PARAM_MISMATCH;
		pthread_mutex_unlock(&this->ptp_mtx);
		return(0);
	}

	// wait for answer from ptpclient
	do {
		len = this->ptp_raw.recv_ptp(this->rtt * 2, &packet);
		if(!len) {
			this->error_code = PTP_ERR_RECVRAW;
			pthread_mutex_unlock(&this->ptp_mtx);
			return(0);
		}

		ip_hdr = (struct iphdr*)packet;
		ptp_hdr = (struct ptphdr*)(packet + ip_hdr->ihl*4);
		data_len = len - ip_hdr->ihl*4 - sizeof(struct ptphdr);

		// the data_len check is '>=' to handle returned echo packets which got the original destination attached
		if(!ptp_hdr->compare && ptp_hdr->echo && (ptp_hdr->type == PTP_TYPE_IFMTU) && (ptp_hdr->length == 2) && (data_len >= (signed)sizeof(__u32))) {
			mtu = ntohl(*((__u32 *)((unsigned char *)ptp_hdr + sizeof(struct ptphdr))));
			*missing = ptp_hdr->count;
			delete(packet);
			pthread_mutex_unlock(&this->ptp_mtx);
			return(mtu);
		}

		delete(packet);
	}
	while(1);
}

__u32 PTPserver::get_ifspeed(int compare, int echo, __u32 min_ifspeed, __u8 expected_hops, __u8 *missing)
{
	int		len, data_len;
	__u32		ifspeed, min_ifspeed_n;
	unsigned char	*packet, *padding;
	struct iphdr	*ip_hdr;
	struct ptphdr	*ptp_hdr;

	pthread_mutex_lock(&this->ptp_mtx);
	this->error_code = PTP_ERR_NONE;

	if(!compare && !echo) {
		if(expected_hops > 0) {
			padding = new(unsigned char[expected_hops * 4]);
			memset(padding, 0, expected_hops * 4);
		}
		else
			padding = NULL;

		packet = this->ptp_raw.assemble_ptp(PTP_DEFAULT_TTL, this->destination.s_addr, PTP_DEFAULT_TTL, compare, echo, 2,0, PTP_TYPE_IFSPEED, expected_hops * 4, 0, padding, &len);

		if(padding)
			delete(padding);

		if(packet) {
			if(!this->ptp_raw.send_ptp(len, packet)) {
				delete(packet);
				this->error_code = PTP_ERR_SENDRAW;
				pthread_mutex_unlock(&this->ptp_mtx);
				return(0);
			}

			delete(packet);
		}
		else {
			this->error_code = PTP_ERR_CREATERAW;
			pthread_mutex_unlock(&this->ptp_mtx);
			return(0);
		}
	}
	else if(compare) {
		min_ifspeed_n = htonl(min_ifspeed);
		packet = this->ptp_raw.assemble_ptp(PTP_DEFAULT_TTL, this->destination.s_addr, PTP_DEFAULT_TTL, compare, echo, 2,1, PTP_TYPE_IFSPEED, 4, 4, (unsigned char *)&min_ifspeed_n, &len);
		if(packet) {
			if(!this->ptp_raw.send_ptp(len, packet)) {
				delete(packet);
				this->error_code = PTP_ERR_SENDRAW;
				pthread_mutex_unlock(&this->ptp_mtx);
				return(0);
			}

			delete(packet);
		}
		else {
			this->error_code = PTP_ERR_CREATERAW;
			pthread_mutex_unlock(&this->ptp_mtx);
			return(0);
		}
	}
	else {
		this->error_code = PTP_ERR_PARAM_MISMATCH;
		pthread_mutex_unlock(&this->ptp_mtx);
		return(0);
	}

	// wait for answer from ptpclient
	do {
		len = this->ptp_raw.recv_ptp(this->rtt * 2, &packet);
		if(!len) {
			this->error_code = PTP_ERR_RECVRAW;
			pthread_mutex_unlock(&this->ptp_mtx);
			return(0);
		}

		ip_hdr = (struct iphdr*)packet;
		ptp_hdr = (struct ptphdr*)(packet + ip_hdr->ihl*4);
		data_len = len - ip_hdr->ihl*4 - sizeof(struct ptphdr);

		// the data_len check is '>=' to handle returned echo packets which got the original destination attached
		if(!ptp_hdr->compare && ptp_hdr->echo && (ptp_hdr->type == PTP_TYPE_IFSPEED) && (ptp_hdr->length == 2) && (data_len >= (signed)sizeof(__u32))) {
			ifspeed = ntohl(*((__u32 *)((unsigned char *)ptp_hdr + sizeof(struct ptphdr))));
			*missing = ptp_hdr->count;
			delete(packet);
			pthread_mutex_unlock(&this->ptp_mtx);
			return(ifspeed);
		}

		delete(packet);
	}
	while(1);
}

int PTPserver::start_availbw_detection(int type, __u8 expected_hops)
{
	this->stop_availbw_detection();

	if((type != PTP_BANDWIDTH_TYPE_3) && (type != PTP_BANDWIDTH_TYPE_4)) {
		this->error = PTP_ERR_PARAM_MISMATCH;
		return(0);
	}

	this->type = type;
	this->hops = expected_hops;
	this->bandwidth_rtt = this->rtt;

	pthread_attr_init(&this->ptp_attr);

	if(pthread_create(&this->ptp_thread, &this->ptp_attr, ptp_thread_stub, (void *)this)) {
		// error creating thread
		this->error = PTP_BANDWIDTH_ERR_CREATETHREAD;
		return(0);
	}

	this->interval = 0;
	this->error = PTP_ERR_NONE;
	this->running = 1;
	return(1);
}

void PTPserver::stop_availbw_detection(void)
{
	if(this->running) {
		this->running = 0;	// this will cause the thread to set error to PTP_ERR_STOPPED when he's done

		while(this->error != PTP_BANDWIDTH_ERR_STOPPED)
			usleep(1000*50);	// 50ms

		//pthread_cancel(this->ptp_thread);
		pthread_attr_destroy(&this->ptp_attr);
	}
}

void PTPserver::thread_run(void)
{
	int		init_route, len, data_len, done_recv;
	unsigned char	*packet, *padding;
	struct iphdr	*ip_hdr;
	struct ptphdr	*ptp_hdr;
	struct ptpmsg	*ptp_msg;
	clock_t		time_start, time_end, time_lost;
	struct tms	my_tms;
	__u32		temp_rci;

	// flush any queued ptp packets
	pthread_mutex_lock(&this->ptp_mtx);
	while(this->ptp_raw.recv_ptp(0, &packet))
		delete(packet);
	pthread_mutex_unlock(&this->ptp_mtx);

	/* Use TYPE 4 instead of TYPE 3 and check Code of Type 3 before testing */
	if(this->type == PTP_BANDWIDTH_TYPE_4) {
		// send PTP_BANDWIDTH_TYPE_4 packets, no matter what happens

		// init_route == 1 -> send initial packet (to init route and take first snapshot, don't expect a response)
		// init_route == 0 -> send usual traffic measurement packet (expect a response from there on)
		// (we ALWAYS wait for a possible response, it might be an old one which would interfere with the measurement if we wouldn't eat it)
		init_route = 1;
		do {
			time_start = times(&my_tms);

			if(this->hops > 0) {
				padding = new(unsigned char[this->hops * 16]);
				memset(padding, 0, this->hops * 16);
			}
			else
				padding = NULL;

			packet = this->ptp_raw.assemble_ptp(PTP_DEFAULT_TTL, this->destination.s_addr, PTP_DEFAULT_TTL, 0, 0, 8, 0, PTP_TYPE_ADDR_TIME_IFSPEED_IFOCTETS, this->hops * 16, 0, (unsigned char *)padding, &len);

			if(padding)
				delete(padding);

			pthread_mutex_lock(&this->ptp_mtx);

			if(packet) {
				_DMSG0("[thread_run(): sending type 4 packet]\n");
				if(!this->ptp_raw.send_ptp(len, packet)) {
					_DMSG0("[thread_run(): error occured while sending packet]\n");
					this->error = PTP_BANDWIDTH_ERR_SENDRAW;
				}

				delete(packet);
			}
			else {
				_DMSG0("[thread_run(): error assembling type 4 packet]\n");
				this->error = PTP_BANDWIDTH_ERR_CREATERAW;
			}

			done_recv = 0;
			do {
				len = this->ptp_raw.recv_ptp(this->bandwidth_rtt * 2, &packet);

				if(len) {
					ip_hdr = (struct iphdr*)packet;
					ptp_hdr = (struct ptphdr*)(packet + ip_hdr->ihl*4);
					data_len = len - ip_hdr->ihl*4 - sizeof(struct ptphdr);

					if(!ptp_hdr->compare && ptp_hdr->echo && (ptp_hdr->type == PTP_TYPE_ADDR_TIME_IFSPEED_IFOCTETS) && (ptp_hdr->length == 6) && (data_len >= 4)) {
						// an overflow-err or route-changed-err occurred
						done_recv = 1;
						ptp_msg = (struct ptpmsg*)((unsigned char *)ptp_hdr + sizeof(struct ptphdr));

						temp_rci = ntohl(ptp_msg->rci);

						/* bitcontrol if route changed */
						if ((temp_rci & ROUTE_CHANGE) == ROUTE_CHANGE) this->route = 1;
						else this->route = 0;

						/* bitcontrol if there was a counter overflow */
						if ((temp_rci & COUNTER_OVERFLOW) == COUNTER_OVERFLOW) this->counter = 1;
						else this->counter = 0;

						if(this->route) {
							// route changed err msg
							_DMSG0("[thread_run(): received route changed err]\n");
							init_route = 1;
							this->interval = 0;
						}
						else if(this->counter) {
							// counter overflow err msg, adjust thread-RTT
							_DMSG1("[thread_run(): received overflow err, adjusting to interval: %d]\n", temp_rci & (~(COUNTER_OVERFLOW | ROUTE_CHANGE)));
							this->bandwidth_rtt = temp_rci & (~(COUNTER_OVERFLOW | ROUTE_CHANGE));
							this->interval = 0;
						}
						else {
							// New Feedback Msg
							_DMSG1("[thread_run(): received feedback message, traffic = %d]\n", ptp_msg->traffic);
							this->interval = temp_rci & (~(COUNTER_OVERFLOW | ROUTE_CHANGE));
							this->mib2_ifspeed = ptp_msg->mib2_ifspeed;
							this->traffic = ptp_msg->traffic;
						}
						this->error = PTP_BANDWIDTH_ERR_NONE;
					}
					else {
						_DMSG0("[thread_run(): received nonsense packet, skipping]\n");
						delete(packet);
					}
				}
				else {
					// timeout or recv error
					_DMSG1("[thread_run(): error receiving PTP packet (%s)]\n", init_route ? "didn't expect one though, initializing" : "timeout/receive error");
					done_recv = 1;
					//if(init_route)
					//	this->error = PTP_BANDWIDTH_ERR_RECVRAW;
					//else
					this->error = PTP_BANDWIDTH_ERR_NONE;
				}
			} while(!done_recv);

			if(this->error == PTP_BANDWIDTH_ERR_NONE)
				init_route = 0;		// next time we can obtain results for sure

			pthread_mutex_unlock(&this->ptp_mtx);

			if(running) {
				// consume rest of RTT*2 time (this is ok for errors as well) - give the message the max. time do come back
				time_end = times(&my_tms);
				time_lost = (long)(((double)time_end - (double)time_start) / (double)CLOCKS_PER_SEC * (double)1000000);
				if(time_lost < (this->bandwidth_rtt*1000*2))
					usleep((this->bandwidth_rtt*1000*2) - time_lost);

			}
		} while(this->running);

		this->error = PTP_BANDWIDTH_ERR_STOPPED;
	}
	else if (this->type == PTP_BANDWIDTH_TYPE_3) {
		// PTP_BANDWIDTH_TYPE_3

		// send type 2 packet to initialize route, then type 3 packets until a route-change error message appears

		// init_route == 2 -> send initial route-discovery packet (to init route, don't expect a response)
		// init_route == 1 -> send first traffic measurement packet (to take first snapshot, don't expect a response)
		// init_route == 0 -> send usual traffic measurement packet (expect a response from there on)
		// (we ALWAYS wait for a possible response, it might be an old one which would interfere with the measurement if we wouldn't eat it)
		init_route = 2;
		do {
			time_start = times(&my_tms);

			if(this->hops > 0) {
				padding = new(unsigned char[this->hops * 12]);
				memset(padding, 0, this->hops * 12);
			}
			else
				padding = NULL;

			if (init_route == 2) {
				// prepare route initialization packet (type 2)
				packet = this->ptp_raw.assemble_ptp(PTP_DEFAULT_TTL, this->destination.s_addr, PTP_DEFAULT_TTL, 0, 0, 8, 0, PTP_TYPE_ADDR_IFSPEED, this->hops * 8, 0, (unsigned char *)padding, &len);
			}
			else {
				// prepare traffic packet (type 3)
				packet = this->ptp_raw.assemble_ptp(PTP_DEFAULT_TTL, this->destination.s_addr, PTP_DEFAULT_TTL, 0, 0, 8, 0, PTP_TYPE_ADDR_TIME_IFOCTETS, this->hops * 12, 0, (unsigned char *)padding, &len);
			}

			if(padding)
				delete(padding);

			pthread_mutex_lock(&this->ptp_mtx);

			if(packet) {
				_DMSG1("[thread_run(): sending type %d packet]\n", (init_route == 2) ? 2 : 3);
				if(!this->ptp_raw.send_ptp(len, packet)) {
				_DMSG0("[thread_run(): error occured while sending packet]\n");
				this->error = PTP_BANDWIDTH_ERR_SENDRAW;
				}
				else {
					if(init_route)
						this->error = PTP_BANDWIDTH_ERR_NONE;
				}

				delete(packet);
			}
			else {
				_DMSG1("[thread_run(): error assembling type %d packet]\n", (init_route == 2) ? 2 : 3);
				this->error = PTP_BANDWIDTH_ERR_CREATERAW;
			}

			done_recv = 0;
			do {
				len = this->ptp_raw.recv_ptp(this->bandwidth_rtt * 2, &packet);
				if(len) {
					ip_hdr = (struct iphdr*)packet;
					ptp_hdr = (struct ptphdr*)(packet + ip_hdr->ihl*4);
					data_len = len - ip_hdr->ihl*4 - sizeof(struct ptphdr);

					if(!ptp_hdr->compare && ptp_hdr->echo && (ptp_hdr->type == PTP_TYPE_ADDR_TIME_IFOCTETS) && (ptp_hdr->length == 2) && (data_len >= 4)) {
						// an overflow-err or route-changed-err occurred
						done_recv = 1;
						ptp_msg = (struct ptpmsg*)((unsigned char *)ptp_hdr + sizeof(struct ptphdr));

						temp_rci = ntohl(ptp_msg->rci);

						/* bitcontrol if route changed */
						if ((temp_rci & ROUTE_CHANGE) == ROUTE_CHANGE) this->route = 1;
						else this->route = 0;

						/* bitcontrol if there was a counter overflow */
						if ((temp_rci & COUNTER_OVERFLOW) == COUNTER_OVERFLOW) this->counter = 1;
						else this->counter = 0;

						if(this->route) {
								// route changed err msg (or the initial route-discovery packet got lost)
							_DMSG0("[thread_run(): received route changed err]\n");
							init_route = 2;
							this->interval = 0;
						}
						else if(this->counter) {
							// counter overflow err msg, adjust thread-RTT
							_DMSG1("[thread_run(): received overflow err, adjusting to interval: %d]\n", temp_rci & (~(COUNTER_OVERFLOW | ROUTE_CHANGE)));
							this->bandwidth_rtt = temp_rci & (~(COUNTER_OVERFLOW | ROUTE_CHANGE));
							this->interval = 0;
						}
						else {
							// new bandwidth msg
							_DMSG1("[thread_run(): received new bandwidth: %d]\n", ptp_msg->mib2_ifspeed);
							this->interval = ntohl(temp_rci & (~(COUNTER_OVERFLOW | ROUTE_CHANGE)));
							this->mib2_ifspeed = ptp_msg->mib2_ifspeed;
							this->traffic = ptp_msg->traffic;
						}
						this->error = PTP_BANDWIDTH_ERR_NONE;
					}
					else {
						_DMSG0("[thread_run(): received nonsense packet, skipping]\n");
						delete(packet);
					}
				}
				else {
					// timeout or recv error
					_DMSG1("[thread_run(): error receiving PTP packet (%s)]\n", init_route ? "didn't expect one though, initializing" : "timeout/receive error");
					done_recv = 1;
					if(!init_route)
						this->error = PTP_BANDWIDTH_ERR_RECVRAW;
					else
						this->error = PTP_BANDWIDTH_ERR_NONE;
				}
			} while(!done_recv);

			if((this->error == PTP_BANDWIDTH_ERR_NONE) && init_route)
				init_route--;

			pthread_mutex_unlock(&this->ptp_mtx);

			if(running) {
				// consume rest of RTT*2 time (this is ok for errors as well)
				time_end = times(&my_tms);
				time_lost = (long)(((double)time_end - (double)time_start) / (double)CLOCKS_PER_SEC * (double)1000000);
				if(time_lost < (this->bandwidth_rtt*1000*2))
					usleep((this->bandwidth_rtt*1000*2) - time_lost);

			}
		} while(this->running);

		this->error = PTP_BANDWIDTH_ERR_STOPPED;
	}
}

int PTPserver::get_bandwidth(int *route, int *counter, __u32 *interval, __u32 *mib2_ifspeed, __u32 *traffic)
{
	*route = this->route;
	*counter = this->counter;
	*interval = this->interval;
	*mib2_ifspeed = this->mib2_ifspeed;
	*traffic = this->traffic;

	this->error  = PTP_BANDWIDTH_ERR_NONE;

	return(this->error);
}
