#ifndef	__PTPSERVER_H
#define	__PTPSERVER_H

#include <linux/types.h>
#include <pthread.h>
#include <netinet/in.h>
#include "ptpcommon.h"
#include "ptpraw.h"

// the various error codes which can be returned by get_error()
#define	PTP_ERR_NONE		1000
#define	PTP_ERR_HOSTLOOKUP	1001
#define	PTP_ERR_CREATESOCKET	1002
#define	PTP_ERR_NOSUCHDEVICE	1003
#define	PTP_ERR_INITRAW		1004
#define	PTP_ERR_CREATERAW	1005
#define	PTP_ERR_RECVRAW		1006
#define	PTP_ERR_SENDRAW		1007
#define	PTP_ERR_PARAM_MISMATCH	1008
#define	PTP_ERR_TIMEOUT		1009

// bandwidth measurement values
#define	PTP_BANDWIDTH_TYPE_3	0
#define	PTP_BANDWIDTH_TYPE_4	1

// the various error codes which can be set by get_bandwidth()
#define	PTP_BANDWIDTH_ERR_NONE		2000
#define	PTP_BANDWIDTH_ERR_STOPPED		2001
#define	PTP_BANDWIDTH_ERR_CREATETHREAD	2002
#define	PTP_BANDWIDTH_ERR_PARAM_MISMATCH	2003
#define	PTP_BANDWIDTH_ERR_CREATERAW		2004
#define	PTP_BANDWIDTH_ERR_RECVRAW		2005
#define	PTP_BANDWIDTH_ERR_SENDRAW		2006
#define	PTP_BANDWIDTH_ERR_TIMEOUT		2007

class PTPserver {
private:
	int		is_init, running, error_code, error, type;
	__u8		hops;
	__u16		ptp_port;
	int route, counter;
	__u32		interval, mib2_ifspeed, traffic;
	unsigned long	device;
	long		rtt, bandwidth_rtt;
	struct in_addr	destination;
	pthread_t	ptp_thread;
	pthread_attr_t	ptp_attr;
	pthread_mutex_t	ptp_mtx;
	PTPraw		ptp_raw;

	int convert_to_inaddr(const char *host, struct in_addr *address);

public:
	PTPserver();
	~PTPserver();

	// init() is the first function to call, regardless if you are going to
	// measure the bandwidth with the thread, or if you use non-threaded measurement functions
	// rtt is an average starting rtt in milliseconds, the timeout for responses is 2*rtt
	// (this value gets self-adjusted after this class sent and received a few packets [NOT IMPLEMENTED YET])
	// ptp_port is usually set to PTP_DEFAULT_PORT (0)
	// be careful NOT TO CHOOSE the SAME destination AND ptp_port multiple times (thats just like TCP: you cant have
	// two jobs listening on the same device AND port)
	// returns nonzero on success.
	// (multiple calls will issue a stop_availbw_detection() if the bandwidth measurement is running)
	int init(const char *device, const char *destination, long rtt, __u16 ptp_port);

	// returns the most recent error code set by non-threaded functions
	int get_error(void) { return(this->error_code); };

	// non-threaded path MTU
	// compare=0, echo=0: PTP-routers add datasets, receiver will pick the worst MTU and returns it (min_mtu ignored)
	// compare=1, echo=0: PTP-routers update the requested min_mtu with the worst value, receiver returns it
	// compare=1, echo=1: PTP-router will echo the packet as soon as its MTU is worse than the requested one, else receiver returns it
	// the 'expected_hops' value is only used with compare=echo=0 and will fire off packets padded with the correct
	// size of PDU (so that expected_hops times a dataset can fit into the packet and PTP-routers dont need to copy
	// packet contents all the time, before adding a dataset)
	// missing will be set to the number of non-PTP routers upon success
	// returns nonzero on success
	__u32 get_path_mtu(int compare, int echo, __u32 min_mtu, __u8 expected_hops, __u8 *missing);

	// non-threaded ifSpeed
	// compare=0, echo=0: PTP-routers add datasets, receiver will pick the worst ifspeed and returns it (min_ifspeed ignored)
	// compare=1, echo=0: PTP-routers update the requested min_ifspeed with the worst value, receiver returns it
	// compare=1, echo=1: PTP-router will echo the packet as soon as its ifspeed is worse than the requested one, else receiver returns it
	// the 'expected_hops' value is only used with compare=echo=0 and will fire off packets padded with the correct
	// size of PDU (so that expected_hops times a dataset can fit into the packet and PTP-routers dont need to copy
	// packet contents all the time, before adding a dataset)
	// missing will be set to the number of non-PTP routers upon success
	// returns nonzero on success
	__u32 get_ifspeed(int compare, int echo, __u32 min_ifspeed, __u8 expected_hops, __u8 *missing);



	// start_availbw_detection() launches the bandwidth measurement thread
	// you can obtain the current bandwith (and recent error code) with get_bandwith(),
	// and stop measurement with stop_availbw_detection()
	// type can be either PTP__TYPE_3 or PTP__TYPE_4
	// (if the rtt is too high, it will silently adopt to what the receiver tells it to use)
	// this one is independent from the non-threaded functions, you can use both at the same time
	// (packets don't interfere with each other)
	// the 'expected_hops' value is used to fire off packets padded with the correct
	// size of PDU (so that expected_hops times a dataset can fit into the packet and PTP-routers dont need to copy
	// packet contents all the time, before adding a dataset)
	// returns nonzero on success (else use get_bandwith() to obtain the launch error)
	// (multiple calls will issue a stop_availbw_detection() if the bandwidth measurement is running)
	int start_availbw_detection(int type, __u8 expected_hops);

	// get_bandwith() returns the current status of the measurement, if its anything else than PTP__ERR_NONE, don't use
	// the other values.
	// 'route is set to 1 if there was a route change
	// 'counter' is set to 1 if there was a counter overflow (in this case 'interval' contains the max. value
	// 'interval' to the measurement interval used to determine the bandwith
	// 'mib2_ifspeed' contains the current bandwith
	// 'traffic' contains the actual PTP traffic on this route
	int PTPserver::get_bandwidth(int *route, int *counter, __u32 *interval, __u32 *mib2_ifspeed, __u32 *traffic);


	// stop_availbw_detection() terminates the measure thread (this can take a few ms)
	void stop_availbw_detection(void);

	// never call this (must be public for thread workaround)
	void thread_run(void);
};

#endif
